/*
MIT License

Copyright (c) 2021 Sulako

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/


#include "HIDDevice.h"


HIDDevice::HIDDevice(const uint8_t* HIDReport_, unsigned int HIDReportLength_, int8_t ReportID_) : PluggableUSBModule(2, 1, endpointTypes) {
  HIDReport=HIDReport_;
  HIDReportLength=HIDReportLength_;
  ReportID=ReportID_;

  endpointTypes[0] = EP_TYPE_BULK_IN;
  endpointTypes[1] = EP_TYPE_BULK_OUT;
  PluggableUSB().plug(this);
}

int HIDDevice::getInterface(uint8_t *interfaceNumber) {
  interfaceNumber[0] += 1;
  HIDDescriptorInOut interfaceDescriptor = {
    D_INTERFACE(pluggedInterface, 2, USB_DEVICE_CLASS_HUMAN_INTERFACE, 0, 0),
    D_HIDREPORT(HIDReportLength),
    D_ENDPOINT(USB_ENDPOINT_IN(pluggedEndpoint), USB_ENDPOINT_TYPE_INTERRUPT, USB_EP_SIZE, 0),
    D_ENDPOINT(USB_ENDPOINT_OUT(pluggedEndpoint+1), USB_ENDPOINT_TYPE_INTERRUPT, USB_EP_SIZE, 0),
  };
  return USB_SendControl(0, &interfaceDescriptor, sizeof(interfaceDescriptor));
}

int HIDDevice::getDescriptor(USBSetup &setup) {
  if (setup.bmRequestType != REQUEST_DEVICETOHOST_STANDARD_INTERFACE) { return 0; }
  if (setup.wValueH != HID_REPORT_DESCRIPTOR_TYPE) { return 0; }
  if (setup.wIndex != pluggedInterface) { return 0; }

  protocol = HID_REPORT_PROTOCOL;

  return USB_SendControl(TRANSFER_PGM, HIDReport, HIDReportLength);
}

bool HIDDevice::setup(USBSetup &setup) {

  if (pluggedInterface != setup.wIndex) {
    return false;
  }

  uint8_t request = setup.bRequest;
  uint8_t requestType = setup.bmRequestType;

  if (requestType == REQUEST_DEVICETOHOST_CLASS_INTERFACE)
  {
    if (request == HID_GET_REPORT) {
      return true;
    }
    if (request == HID_GET_PROTOCOL) {
      return true;
    }
  }

  if (requestType == REQUEST_HOSTTODEVICE_CLASS_INTERFACE)
  {
    if (request == HID_SET_PROTOCOL) {
      protocol = setup.wValueL;
      return true;
    }
    if (request == HID_SET_IDLE) {
      idle = setup.wValueL;
      return true;
    }
  }
  return false;
}


int HIDDevice::SendReport(const void* data, int len)
{
  if (ReportID!=0)
  {
    auto ret = USB_Send(pluggedEndpoint, &ReportID, 1);
    if (ret < 0) return ret;
  }
  auto ret2 = USB_Send(pluggedEndpoint | TRANSFER_RELEASE, data, len);
  if (ret2 < 0) return ret2;
  return  ret2;
}

bool HIDDevice::isAvailable()
{
  return USB_Available(pluggedEndpoint+1);
}

unsigned int HIDDevice::receive(uint8_t* data, unsigned int maxlen) {
  unsigned int len=0;
  if (USB_Available(pluggedEndpoint+1)) 
  {
      while (USB_Available(pluggedEndpoint+1) && len<maxlen) {
        data[len]=USB_Recv(pluggedEndpoint+1);
        len++;
      }
      return len;
  }
  else {
    return 0;
  }
}


uint8_t HIDDevice::getShortName(char *name)
{
  name[0] = '0';
  name[1] = '0';
  name[2] = '0';
  name[3] = '0';
  name[4] = '1';
  return 5;
}
