/*
MIT License

Copyright (c) 2021 Sulako

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include "ffb.h"

//Класс для обработки команд ОС и расчета прилагаемой силы
ForceFeedback::ForceFeedback()
{
    uint8_t data[]={0xFE,0x0D,0x06,0x06,0xb4,0,0}; //Данные дефолтной пружины

    forceSlots[0]=new FFB_Spring(data, 0);
    forceSlots[1]=new FFB_Force();
    forceSlots[2]=new FFB_Force();
    forceSlots[3]=new FFB_Force();
    forceSlots[4]=new FFB_Force();

    prevAxis=128;
    fixedLoop=true;
   
    wheelRange=270;
    
}

//Обработка команд
void ForceFeedback::processData(uint8_t *data)
{
    /*
     * Есть 4 слота эффектов + дефолтная пружина
     * в любой слот можно поставить любой эффект
     */
  
    uint8_t slots,slot=0;
    uint8_t cmd;

    slots=data[0] >> 4;
    cmd=data[0] & 0b00001111;

    if (data[0]==0xf8)  //Extended Commands
       {
       
            //Задание диапазона руля никак не обрабатывается
            switch (data[1])
            {
                case 0x02:
                    wheelRange=200; 
                    break;
                case 0x03:
                    wheelRange=900;
                    break;
                case 0x81:
                    wheelRange=((unsigned int)data[3]<<8)|data[2];
                    if (wheelRange>900)
                        wheelRange=900;
                    if (wheelRange<40)
                        wheelRange=40;
                    break; 
                    
            }
      
            
        }
    else
    if (cmd==0x0d)//Fixed loop
    {
        if (data[1])
                fixedLoop=true;
            else
                fixedLoop=false;
    }
    else
    {
        for(slot=1;slot<=4;slot++)
        if ((1<<(slot-1))&slots)
        {
            switch (cmd)
            {
                case 0x00: //Download Force
                    downloadForce(data,slot);
                    break;
                 case 0x01: //Download and Play Force
                    downloadForce(data,slot);
                    forceSlots[slot]->active=true;
                    break;
                case 0x02: //Play Force
                    forceSlots[slot]->active=true;
                    break;
                case 0x03: //Stop Force
                    forceSlots[slot]->active=false;
                    break;
                case 0x04://Default Spring On
                    if (slot==1)
                        forceSlots[0]->active=true;
                    break;
                case 0x05://Default Spring Off
                    if (slot==1)
                        forceSlots[0]->active=false;
                    break;
                case 0x0c: //Refresh Force
                    forceSlots[slot]->update(data,slot);
                    break;
                case 0x0e: //set default spring
                    if (slot==1)
                        downloadForce(data,0);
                    break;
            }
        }
    }
}

//Установка эффекта в слот
void ForceFeedback::downloadForce(uint8_t *data, uint8_t slot)
{
        if (forceSlots[slot]!=NULL)
            delete forceSlots[slot];

        switch (data[1])
        {
            case 0x00:
                forceSlots[slot]=new FFB_Constant(data, slot);
                break;
            case 0x01:
            case 0x03:
            case 0x0b:
            case 0x0d:
                forceSlots[slot]=new FFB_Spring(data, slot);
                break;
            case 0x02:
            case 0x0c:
            case 0x0e:
                forceSlots[slot]=new FFB_Damper(data, slot);
                break;
            case 0x04:
            case 0x05:
                forceSlots[slot]=new FFB_Sawtooth(data, slot);
                break;
            case 0x06:
                forceSlots[slot]=new FFB_Trapezoid(data, slot);
                break;
            case 0x07:
                forceSlots[slot]=new FFB_Rectangle(data, slot);
                break;
            case 0x08:
                forceSlots[slot]=new FFB_Variable(data, slot);
                break;
            case 0x09:
                forceSlots[slot]=new FFB_Ramp(data, slot);
                break;
            default:
                forceSlots[slot]=NULL;
        }

}

//Расчет силы
int8_t ForceFeedback::getValue(unsigned int axis)
{
    /*
        Каждая сила возвращает от -127 до 127.
        Складываем все вместе и обрезаем на максимуме 127.

        также, блокируем силу при достижении крайних положений чтобы не сломать чего-нибудь.
    */
    
    uint8_t s;
    int value=0;

    for(s=0;s<=4;s++)
        if (forceSlots[s]!=NULL)
        if (forceSlots[s]->active)
            value+=forceSlots[s]->getValue(axis);

    value=constrain(value, -127, 127);

    if ((axis<500)&&(value<0))
        value=0;
    if ((axis>=60000)&&(value>0))
        value=0;

    return value;
}

//Для отладки
void ForceFeedback::info()
{
    uint8_t s;
    for(s=0;s<=4;s++)
    {
        Serial.print("["+String(s)+"]: ");
        if (forceSlots[s]==NULL)
            Serial.print("none");
        else
        {
            if (forceSlots[s]->active)
                Serial.print("Playing ");
            forceSlots[s]->info();
        }
        Serial.println();
    }
    Serial.println("fixedLoop: "+String(fixedLoop));
    Serial.println("wheelRange: "+String(wheelRange));

}
