/*
MIT License

Copyright (c) 2021 Sulako 

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include "ffb_forces.h"

//Различные эффекты сил

//Коэффициенты для пружин и демпфера
uint8_t KFactor(uint8_t offset, uint8_t K, bool hires)
{
    unsigned int force=0;

    if (hires)
         force=offset*(K+1)/4;
    else
    switch (K)
        {
            case 0x00:
                force=offset/4;
                break;
            case 0x01:
                force=offset/2;
                break;
            case 0x02:
                force=offset*3/4;
                break;
            case 0x03:
                force=offset;
                break;
            case 0x04:
                force=offset*3/2;
                break;
            case 0x05:
                force=offset*2;
                break;
            case 0x06:
                force=offset*3;
                break;
            case 0x07:
                force=offset*4;
                break;
        }
    if (force>127)
        force=127;
    return force;
}

/*
    В протоколе усилия кодируются от 0 до 255 где 127-128 нулевое усилие
    Это неудобно, поэтому здесь на выходе у сил будет диапазон -127..0..127
*/
int8_t convertForce(uint8_t f)
{
    if (f>=128)
        return 128-f;
    else
        return 127-f;
}

FFB_Spring::FFB_Spring(uint8_t* data, uint8_t slot)
{
    hires=false;
    update(data,slot);
}
void FFB_Spring::update(uint8_t* data, uint8_t slot)
{
    if ((data[1]==0x01)||(data[1]==0x0b))
    {
        D1=data[2];
        D2=data[3];
        K1=data[4]&0b00001111;
        K2=data[4]>>4;
        S1=data[5]&0b00000001;
        S2=data[5]&0b00010000;
        Clip=data[6];
    }
    if ((data[1]==0x00)||(data[1]==0x03)||(data[1]==0x0d))
    {
        D1=126;
        D2=129;
        K1=data[2]&0b00001111;
        K2=data[3]&0b00001111;
        S1=0;
        S2=0;
        Clip=data[4];
    }

    /*  Отступления от протокола:
        1) границы мертвой зоны D1 и D2 драйвер всегда устанавливает в 0 (левый максимум), что не имеет смысла.
        поэтому в этом случае считаем что это центрующие пружины
        2) условное значение силы в любую сторону по протоколу не более 127, но ограничитель Clip ставится от 0 до 255.
        так что масштабируем.
    */
    if ((!D1)&&(!D2))
    {
        D1=127;
        D2=128;
    }
    Clip=Clip/2;


    if ((data[1]==0x0b)||(data[1]==0x0d))
        hires=true;
}
int8_t FFB_Spring::getValue(unsigned int axis)
{
    uint8_t offset;
    unsigned int force;
    int8_t K,sign;

    axis=axis>>8;

    if (axis>D2)
    {
        offset=axis-D2;
        sign=S2?1:-1;
        K=K2;
    }
    else
    if (axis<D1)
    {
        offset=D1-axis;
        sign=S1?-1:1;
        K=K1;
    }
    else
        return 0;

    force=KFactor(offset, K, hires);

    if (force>Clip)
        force=Clip;

    return sign*force;
}
void FFB_Spring::info()
{
  /*mySerial.print("Spring: D1="+String(D1)+
                      " D2="+String(D2)+
                      " K1="+String(K1)+
                      " K2="+String(K2)+
                      " S1="+String(S1)+
                      " S2="+String(S2)+
                      " Clip="+String(Clip)+
                      " hires="+String(hires));
     */
}


FFB_Damper::FFB_Damper(uint8_t* data, uint8_t slot){
    hires=false;
    friction=false;
    update(data,slot);
}
void FFB_Damper::update(uint8_t* data, uint8_t slot)
{
    if ((data[1]==0x02)||(data[1]==0x0c)) //Damper
    {
        K1=data[2]&0b00001111;
        K2=data[4]&0b00001111;
        S1=data[3]&0b00000001;
        S2=data[5]&0b00000001;
        if (data[1]==0x0c)
            hires=true;
    }
    if (data[1]==0x0e) //Friction
    {
        K1=data[2]&0b00001111;
        K2=data[3]&0b00001111;
        Clip=data[4];
        S2=data[5]&0b00010000;
        S1=data[5]&0b00000001;
        friction=true;
        Clip=Clip/2; //масштабирование
    }
    flagFirstRun=true;
    value=0;
    loops=0;
}
int8_t FFB_Damper::getValue(unsigned int axis)
{
    /*  по протоколу расчет сил делается 500 раз в секунду (период 2мс) или чаще, если выключен Fixed Loop
     *  демпфер рассчитывает усилие исходя из смещения руля с прошлого расчета
     *  положение руля также в протоколе задается 1 байтом - 0..255
     *  Если диапазон руля 900 градусов то приходится по 3.5 градуса на отсчет.
     *  Чтобы сработал демпфер, минимальная скорость вращения руля должна быть не менее чем 1 отсчет за 2мс, 
     *  или 500 отсчетов за 1сек, или 1757 градусов/сек, или 4.9 оборотов/сек
     *  ерунда какая-то, руль невозможно повернуть так быстро
     *  поэтому чтобы демпфер вообще срабатывал, расчет делаем в 10 раз реже. 
     */
    unsigned int offset;
    unsigned int force;
    int8_t K,sign;

    loops++;
    if (loops<10)
      return value;
      
    loops=0;
       
    axis=axis>>8;
    if (flagFirstRun)
    {
      flagFirstRun=false;
      prevAxis=axis;
    }

    if (prevAxis<axis)
    {
        offset=axis-prevAxis;
        sign=S2?1:-1;
        K=K2;
    }
    else
    if (prevAxis>axis)
    {
        offset=prevAxis-axis;
        sign=S1?-1:1;
        K=K1;
    }
    else
        return 0;

        
    prevAxis=axis;    
 
    force=KFactor(offset,K,hires);

    /*
     * Эффект трения по расчету аналогичен демпферу, 
     * но противодействующее усилие начинается сразу с некоторого значения 
     * и уменьшается по мере увеличения скорости вращения
     */
    if (friction)         
        force=Clip-force;

    value=sign*force;
   
    
    return value;
}
void FFB_Damper::info()
{
   /* mySerial.print("Damper: K1="+String(K1)+
                        " K2="+String(K2)+
                        " S1="+String(S1)+
                        " S2="+String(S2)+
                        " Clip="+String(Clip)+
                        " hires="+String(hires)+
                        " friction="+String(friction));
                        */
}

FFB_Variable::FFB_Variable(uint8_t* data, uint8_t slot)
{
    update(data, slot);
}
void FFB_Variable::update(uint8_t* data, uint8_t slot)
{
    if (slot==1)
    {
        value=data[2];
        T=data[4]>>4;
        S=data[4]&0b00001111;
        D=data[6]&0b00000001;
    }

    if (slot==3)
    {
        value=data[3];
        T=data[5]>>4;
        S=data[5]&0b00001111;
        D=data[6]&0b00010000;
    }
    /*
    на деле всегда устанавливается только параметр L1, L2 всегда 128, остальные 0
    то есть сила Variable работает как постоянная
    очень странный logitech
    */
}
int8_t FFB_Variable::getValue(unsigned int axis)
{
    if (T&&S)
    {
        if (loops>T)
        {
            if (!D)
            {
                if (value<255-S)
                    value+=S;
                else
                    value=255;
            }
            else
            {
                if (value>S)
                    value-=S;
                else
                    value=0;
            }
            loops=0;
        }
        loops++;
    }

    return convertForce(value);
}
void FFB_Variable::info()
{
   // mySerial.print("Variable: value="+String(value)+" T="+String(T)+" D="+String(D)+" S="+String(S)+" loops="+String(loops));

}
/*
 * Следующие эффекты в играх не замечены, 
 * и с их описаниями не все понятно, но на всякий случай реализация
*/

FFB_Constant::FFB_Constant(uint8_t* data, uint8_t slot)
{
    update(data, slot);
}
void FFB_Constant::update(uint8_t* data, uint8_t slot)
{
    level=convertForce(data[slot+1]);
}
int8_t FFB_Constant::getValue(unsigned int axis)
{
    return level;
}
void FFB_Constant::info()
{
   // mySerial.print("Constant: level="+String(level));
}

FFB_Sawtooth::FFB_Sawtooth(uint8_t* data, uint8_t slot)
{
    update(data, slot);
    loops=0;
}
void FFB_Sawtooth::update(uint8_t* data, uint8_t slot)
{
    L1=data[2];
    L2=data[3];
    Inc=data[6]&0b00001111;
    T3=data[6]>>4;
    value=data[4];

    //коррекция на всякий случай
    if (L1<L2)
    {
        L2=data[2];
        L1=data[3];
    }


    if (data[1]==0x05)
        Inc=-Inc;
}
int8_t FFB_Sawtooth::getValue(unsigned int axis)
{
    if (loops>T3)
    {
        loops=0;

        if (Inc>0)
        {
            if (value<L1-Inc)
                value+=Inc;
            else
                if (value<L1)
                    value=L1;
                else
                    value=L2;
        }
        else
        {
            if (value>L2-Inc)
                value+=Inc;
            else
                if (value>L2)
                    value=L2;
                else
                    value=L1;
        }

    }
    loops++;

    return convertForce(value);
}
void FFB_Sawtooth::info()
{
   // mySerial.print("Sawtooth: L1="+String(L1)+" L2="+String(L2)+" T3="+String(T3)+" Inc="+String(Inc)+" value="+String(value)+" loops="+String(loops));
}

FFB_Trapezoid::FFB_Trapezoid(uint8_t* data, uint8_t slot)
{
    update(data,slot);
    value=L1;
    loops=0;
    mode=0;
}
void FFB_Trapezoid::update(uint8_t* data, uint8_t slot)
{
    L1=data[2];
    L2=data[3];
    T1=data[4];
    T2=data[5];
    T3=data[6]>>4;
    S=data[6]&0b00001111;

    //коррекция на всякий случай
    if (L1<L2)
    {
        L2=data[2];
        L1=data[3];
    }

}
int8_t FFB_Trapezoid::getValue(unsigned int axis)
{
    switch (mode)
    {
        case 0: //спуск
            if (loops>T3)
            {
                loops=0;

                if (value>L2+S)
                    value-=S;
                else
                    if (value>L2)
                        value=L2;
                    else
                        mode=1;
            }
            break;
        case 1: //Полка снизу
            if (loops>T2)
            {
                mode=2;
                loops=0;
            }
            break;
        case 2: //Рост
            if (loops>T3)
            {
                loops=0;

                if (value<L1-S)
                    value+=S;
                else
                    if (value<L1)
                        value=L1;
                    else
                        mode=3;
            }
            break;
        case 3: //Полка сверху
            if (loops>T1)
            {
                mode=0;
                loops=0;
            }
            break;
    }

    loops++;

    return convertForce(value);
}
void FFB_Trapezoid::info()
{
  //  mySerial.print("Trapezoid: L1="+String(L1)+" L2="+String(L2)+" T1="+String(T1)+" T2="+String(T2)+" T3="+String(T3)+" S="+String(S)+" value="+String(value)+" loops="+String(loops)+" mode="+String(mode));
}

FFB_Rectangle::FFB_Rectangle(uint8_t* data, uint8_t slot)
{
    update(data,slot);
    if (P>T1+T2)
        P=0;
    if (P<T1/2)
        value=L2;
    else
        value=L1;

    value=L1;
    loops=0;
}
void FFB_Rectangle::update(uint8_t* data, uint8_t slot)
{
    L1=data[2];
    L2=data[3];
    T1=data[4];
    T2=data[5];
    P=data[6];
}
int8_t FFB_Rectangle::getValue(unsigned int axis)
{
    if ((value==L1)&&(loops>T1))
    {
        value=L2;
        loops=0;
    }
    else
    if ((value==L2)&&(loops>T2))
    {
        value=L1;
        loops=0;
    }

    loops++;
    return convertForce(value);
}
void FFB_Rectangle::info()
{
   // mySerial.print(sprintf("Rectangle: L1=%i L2=%i T1=%i T2=%i P=%i value=%i loops=%i",L1,L2,T1,T2,P,value,loops));
}

FFB_Ramp::FFB_Ramp(uint8_t* data, uint8_t slot)
{
    update(data,slot);
    if (D)
        value=L1;
    else
        value=L2;
    loops=0;
}
void FFB_Ramp::update(uint8_t* data, uint8_t slot)
{
    L1=data[2];
    L2=data[3];
    D=data[4]&0b00000001;
    T=data[5]>>4;
    S=data[5]&0b00001111;

    if (L1<L2)
    {
        L2=data[2];
        L1=data[3];
    }
}
int8_t FFB_Ramp::getValue(unsigned int axis)
{
    if (loops>T)
    {
        if (!D)
        {
            if (value<L1-S)
                value+=S;
            else
                if (value<L1)
                    value=L1;
                else
                    active=false;
        }
        else
        {
            if (value>L2+S)
                value-=S;
            else
                if (value>L2)
                    value=L2;
                else
                    active=false;
        }
        loops=0;
    }
    loops++;
    return convertForce(value);
}
void FFB_Ramp::info()
{
   // mySerial.print(sprintf("Ramp: L1=%i L2=%i T=%i S=%i D=%i value=%i loops=%i",L1,L2,T,S,D,value,loops));
}
